# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from tempest import config
from tempest.lib import decorators
import testtools

from trove_tempest_plugin.tests import constants
from trove_tempest_plugin.tests.scenario import base_actions
from trove_tempest_plugin.tests import utils

CONF = config.CONF


class InstanceActionsMySQLBase(base_actions.TestInstanceActionsBase):
    @classmethod
    def init_db(cls, ip, username=constants.DB_USER,
                password=constants.DB_PASS, database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "CREATE TABLE Persons (ID int, String varchar(255));",
            ]
            db_client.mysql_execute(cmds)

    def insert_data_upgrade(self, ip,
                            username=constants.DB_USER,
                            password=constants.DB_PASS,
                            database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "insert into Persons VALUES (99, 'Upgrade');"
            ]
            db_client.mysql_execute(cmds)

    def verify_data_upgrade(self, ip,
                            username=constants.DB_USER,
                            password=constants.DB_PASS,
                            database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Persons;"
            ret = db_client.mysql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertGreaterEqual(len(rows), 1)

        result = []
        for index in range(len(rows)):
            result.append(dict(zip(keys, rows[index])))
        expected = {'ID': 99, 'String': 'Upgrade'}
        self.assert_single_item(result, **expected)

    def insert_data_before_rebuild(self, ip,
                                   username=constants.DB_USER,
                                   password=constants.DB_PASS,
                                   database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "CREATE TABLE Rebuild (ID int, String varchar(255));",
                "insert into Rebuild VALUES (1, 'rebuild-data');"
            ]
            db_client.mysql_execute(cmds)

    def verify_data_after_rebuild(self, ip,
                                  username=constants.DB_USER,
                                  password=constants.DB_PASS,
                                  database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Rebuild;"
            ret = db_client.mysql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertEqual(1, len(rows))

        actual = dict(zip(keys, rows[0]))
        expected = {'ID': 1, 'String': 'rebuild-data'}
        self.assertEqual(expected, actual)

    def get_db_version(self, ip, username=constants.DB_USER,
                       password=constants.DB_PASS):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306'
        with utils.SQLClient(db_url) as db_client:
            cmd = "SELECT @@GLOBAL.innodb_version;"
            ret = db_client.mysql_execute(cmd)
            return ret.first()[0]

    def get_config_value(self, ip, option, username=constants.DB_USER,
                         password=constants.DB_PASS):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306'
        with utils.SQLClient(db_url) as db_client:
            cmd = f"show variables where Variable_name in ('{option}');"
            ret = db_client.mysql_execute(cmd)
            rows = ret.fetchall()
        self.assertEqual(1, len(rows))
        return int(rows[0][1])


class TestInstanceActionsMySQL(InstanceActionsMySQLBase):
    datastore = 'mysql'

    @decorators.idempotent_id("be6dd514-27d6-11ea-a56a-98f2b3cc23a0")
    @testtools.skipUnless(
        'mysql' in CONF.database.pre_upgrade_datastore_versions,
        'Datastore upgrade is disabled.')
    def test_instance_upgrade(self):
        self.instance_upgrade_test()

    @decorators.idempotent_id("27914e82-b061-11ea-b87c-00224d6b7bc1")
    def test_resize(self):
        self.resize_test()

    @decorators.idempotent_id("8d4d675c-d829-11ea-b87c-00224d6b7bc1")
    @testtools.skipUnless(CONF.database.rebuild_image_id,
                          'Image for rebuild not configured.')
    def test_rebuild(self):
        config_values = {"max_connections": 555}
        self.rebuild_test(config_values)


class TestInstanceActionsMariaDB(InstanceActionsMySQLBase):
    datastore = 'mariadb'

    @decorators.idempotent_id("f7a0fef6-f413-11ea-a950-00224d6b7bc1")
    @testtools.skipUnless(
        'mariadb' in CONF.database.pre_upgrade_datastore_versions,
        'Datastore upgrade is disabled.')
    def test_instance_upgrade(self):
        self.instance_upgrade_test()

    @decorators.idempotent_id("fb89d402-f413-11ea-a950-00224d6b7bc1")
    def test_resize(self):
        self.resize_test()

    @decorators.idempotent_id("ff34768e-f413-11ea-a950-00224d6b7bc1")
    @testtools.skipUnless(CONF.database.rebuild_image_id,
                          'Image for rebuild not configured.')
    def test_rebuild(self):
        config_values = {"max_connections": 555}
        self.rebuild_test(config_values)


class TestInstanceActionsPostgreSQL(base_actions.TestInstanceActionsBase):
    datastore = 'postgresql'
    create_user = False
    enable_root = True

    @classmethod
    def init_db(cls, ip):
        db_url = (f'postgresql+psycopg2://root:{cls.password}@'
                  f'{ip}:5432/postgres')
        with utils.SQLClient(db_url) as db_client:
            cmd = "CREATE DATABASE testdb;"
            db_client.pgsql_execute(cmd)

        db_url = (f'postgresql+psycopg2://root:{cls.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "CREATE TABLE persons (id INT PRIMARY KEY NOT NULL, "
                "string VARCHAR(255));",
            ]
            db_client.pgsql_execute(cmds)

    def insert_data_upgrade(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "insert into Persons VALUES (99, 'Upgrade');"
            ]
            db_client.pgsql_execute(cmds)

    def verify_data_upgrade(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Persons;"
            ret = db_client.pgsql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertGreaterEqual(len(rows), 1)

        result = []
        for index in range(len(rows)):
            result.append(dict(zip(keys, rows[index])))
        expected = {'id': 99, 'string': 'Upgrade'}
        self.assert_single_item(result, **expected)

    def insert_data_before_rebuild(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "CREATE TABLE Rebuild (ID int, String varchar(255));",
                "insert into Rebuild VALUES (1, 'rebuild-data');"
            ]
            db_client.pgsql_execute(cmds)

    def verify_data_after_rebuild(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Rebuild;"
            ret = db_client.pgsql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertEqual(1, len(rows))

        actual = dict(zip(keys, rows[0]))
        expected = {'id': 1, 'string': 'rebuild-data'}
        self.assertEqual(expected, actual)

    def get_db_version(self, ip, username=constants.DB_USER,
                       password=constants.DB_PASS):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/postgres')
        with utils.SQLClient(db_url) as db_client:
            cmd = "SHOW server_version;"
            ret = db_client.pgsql_execute(cmd)
            version = ret.first()[0]

        return version.split(' ')[0]

    def get_config_value(self, ip, option):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/postgres')
        with utils.SQLClient(db_url) as db_client:
            cmd = f"SELECT setting FROM pg_settings WHERE name='{option}';"
            ret = db_client.pgsql_execute(cmd)
            rows = ret.fetchall()

        self.assertEqual(1, len(rows))
        return int(rows[0][0])

    @decorators.idempotent_id("97f1e7ca-f415-11ea-a950-00224d6b7bc1")
    @testtools.skipUnless(
        'postgresql' in CONF.database.pre_upgrade_datastore_versions,
        'Datastore upgrade is disabled.')
    def test_instance_upgrade(self):
        self.instance_upgrade_test()

    @decorators.idempotent_id("9b940c00-f415-11ea-a950-00224d6b7bc1")
    def test_resize(self):
        self.resize_test()

    @decorators.idempotent_id("9ec5dd54-f415-11ea-a950-00224d6b7bc1")
    @testtools.skipUnless(CONF.database.rebuild_image_id,
                          'Image for rebuild not configured.')
    def test_rebuild(self):
        config_values = {"max_connections": 101}
        self.rebuild_test(config_values, config_need_restart=True)



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import time

from oslo_log import log as logging
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
import sqlalchemy
from sqlalchemy import text
from tempest.lib import exceptions

LOG = logging.getLogger(__name__)


def wait_for_removal(delete_func, show_func, *args, **kwargs):
    """Call the delete function, then wait for it to be 'NotFound'

    :param delete_func: The delete function to call.
    :param show_func: The show function to call looking for 'NotFound'.
    :param ID: The ID of the object to delete/show.
    :raises TimeoutException: The object did not achieve the status or ERROR in
                              the check_timeout period.
    :returns: None
    """
    check_timeout = 15
    try:
        delete_func(*args, **kwargs)
    except exceptions.NotFound:
        return

    start = int(time.time())
    LOG.info('Waiting for object to be NotFound')
    while True:
        try:
            show_func(*args, **kwargs)
        except exceptions.NotFound:
            return

        if int(time.time()) - start >= check_timeout:
            message = ('%s did not raise NotFound in %s seconds.' %
                       (show_func.__name__, check_timeout))
            raise exceptions.TimeoutException(message)
        time.sleep(3)


def init_engine(db_url):
    return sqlalchemy.create_engine(db_url)


class SQLClient(object):
    def __init__(self, conn_str):
        self.engine = init_engine(conn_str)

    def conn_execute(self, conn, cmds):
        if isinstance(cmds, str):
            result = conn.execute(text(cmds))
            return result

        for cmd in cmds:
            conn.execute(text(cmd))

    def pgsql_execute(self, cmds, **kwargs):
        try:
            with self.engine.connect() as conn:
                conn.connection.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
                return self.conn_execute(conn, cmds)
        except Exception as e:
            raise exceptions.TempestException(
                'Failed to execute database command %s, error: %s' %
                (cmds, str(e))
            )

    def mysql_execute(self, cmds, **kwargs):
        try:
            with self.engine.begin() as conn:
                return self.conn_execute(conn, cmds)
        except Exception as e:
            raise exceptions.TempestException(
                'Failed to execute database command %s, error: %s' %
                (cmds, str(e))
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.engine.dispose()



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from oslo_log import log as logging
from tempest import config

from trove_tempest_plugin.tests import base as trove_base

LOG = logging.getLogger(__name__)
CONF = config.CONF


class TestInstanceActionsBase(trove_base.BaseTroveTest):
    @classmethod
    def init_db(cls, *args, **kwargs):
        pass

    def insert_data_upgrade(self, *args, **kwargs):
        pass

    def verify_data_upgrade(self, *args, **kwargs):
        pass

    def insert_data_before_rebuild(self, *args, **kwargs):
        pass

    def verify_data_after_rebuild(self, *args, **kwargs):
        pass

    def get_db_version(self):
        pass

    def get_config_value(self, ip, option, **kwargs):
        pass

    @classmethod
    def resource_setup(cls):
        super(TestInstanceActionsBase, cls).resource_setup()

        # Initialize database
        LOG.info(f"Initializing data on {cls.instance_ip}")
        cls.init_db(cls.instance_ip)

    def instance_upgrade_test(self):
        cur_version = self.instance['datastore']['version']
        cfg_versions = CONF.database.pre_upgrade_datastore_versions
        ds_version = cfg_versions.get(self.datastore)
        if not ds_version:
            # Fall back to the instance datastore version. In this case, we are
            # still testing the upgrade API but the datastore version doesn't
            # change actually.
            ds_version = cur_version

        name = self.get_resource_name("pre-upgrade")
        LOG.info(f'Creating instance {name} with datastore version '
                 f'{ds_version} for upgrade')
        instance = self.create_instance(name=name,
                                        datastore_version=ds_version,
                                        create_user=self.create_user)
        self.wait_for_instance_status(instance['id'],
                                      expected_op_status=["HEALTHY"])
        instance = self.client.get_resource(
            "instances", instance['id'])['instance']
        instance_ip = self.get_instance_ip(instance)

        # Insert data before upgrading
        LOG.info(f"Initializing data on {instance_ip} before upgrade")
        self.init_db(instance_ip)
        LOG.info(f"Inserting data on {instance_ip} before upgrade")
        self.insert_data_upgrade(instance_ip)

        new_version = cur_version
        LOG.info(f"Upgrading instance {instance['id']} using datastore "
                 f"{new_version}")
        body = {"instance": {"datastore_version": new_version}}
        self.client.patch_resource('instances', instance['id'], body)
        self.wait_for_instance_status(instance['id'],
                                      expected_op_status=["HEALTHY"])

        LOG.info(f"Getting database version on {instance_ip}")
        actual = self.get_db_version(instance_ip)
        self.assertEqual(new_version, actual)

        LOG.info(f"Verifying data on {instance_ip} after upgrade")
        self.verify_data_upgrade(instance_ip)

        # Delete the new instance explicitly to avoid too many instances
        # during the test.
        self.wait_for_instance_status(instance['id'],
                                      expected_status="DELETED",
                                      need_delete=True)

    def resize_test(self):
        # Resize flavor
        LOG.info(f"Resizing flavor to {CONF.database.resize_flavor_id} for "
                 f"instance {self.instance_id}")
        resize_flavor = {
            "resize": {
                "flavorRef": CONF.database.resize_flavor_id
            }
        }
        self.client.create_resource(f"instances/{self.instance_id}/action",
                                    resize_flavor, expected_status_code=202,
                                    need_response=False)
        self.wait_for_instance_status(self.instance_id,
                                      expected_op_status=["HEALTHY"])

        # Verify Trove flavor
        ret = self.client.get_resource('instances', self.instance_id)
        self.assertEqual(CONF.database.resize_flavor_id,
                         ret['instance']['flavor']['id'])

        # Verify Nova flavor
        params = {
            'all_tenants': True,
            'detail': True,
            'name': self.instance['name']
        }
        servers = self.admin_server_client.list_servers(**params)['servers']
        self.assertEqual(1, len(servers))
        self.assertEqual(CONF.database.resize_flavor_id,
                         servers[0]['flavor']['id'])

        # Resize volume
        LOG.info(f"Resizing volume to 2 for instance {self.instance_id}")
        resize_volume = {
            "resize": {
                "volume": {
                    "size": 2
                }
            }
        }
        self.client.create_resource(f"instances/{self.instance_id}/action",
                                    resize_volume, expected_status_code=202,
                                    need_response=False)
        self.wait_for_instance_status(self.instance_id,
                                      expected_op_status=["HEALTHY"])

        # Verify Trove volume
        ret = self.client.get_resource('instances', self.instance_id)
        self.assertEqual(2, ret['instance']['volume']['size'])

    def rebuild_test(self, config_values, config_need_restart=False):
        LOG.info(f"Inserting data on {self.instance_ip} before rebuilding")
        self.insert_data_before_rebuild(self.instance_ip)

        # Create configuration before rebuild
        config_name = self.get_resource_name('config')
        LOG.info(f"Creating new configuration {config_name} for rebuild")
        config = self.create_config(
            config_name, config_values, self.datastore,
            self.instance['datastore']['version'])
        config_id = config['configuration']['id']
        self.addCleanup(self.client.delete_resource, 'configurations',
                        config_id, ignore_notfound=True)
        # Attach the configuration
        LOG.info(f"Attaching config {config_id} to instance "
                 f"{self.instance_id}")
        self.attach_config(self.instance_id, config_id)
        self.addCleanup(self.detach_config, self.instance_id)
        if config_need_restart:
            LOG.info(f"Restarting instance {self.instance_id}")
            self.restart_instance(self.instance_id)
        # Verify the config before rebuild
        key = list(config_values.keys())[0]
        value = list(config_values.values())[0]
        LOG.info(f"Getting config value for {key} on {self.instance_ip}")
        cur_value = self.get_config_value(self.instance_ip, key)
        self.assertEqual(value, cur_value)

        LOG.info(f"Rebuilding instance {self.instance_id} with image "
                 f"{CONF.database.rebuild_image_id}")
        self.rebuild_instance(self.instance_id, CONF.database.rebuild_image_id)

        LOG.info(f"Verifying data on {self.instance_ip} after rebuilding")
        self.verify_data_after_rebuild(self.instance_ip)

        # Verify configuration before rebuild
        LOG.info(f"Verifying config {key} on {self.instance_ip} after "
                 f"rebuilding")
        cur_value = self.get_config_value(self.instance_ip, key)
        self.assertEqual(value, cur_value)



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from oslo_log import log as logging
from tempest.lib import decorators

from trove_tempest_plugin.tests.scenario import base_basic
from trove_tempest_plugin.tests import utils

LOG = logging.getLogger(__name__)


class TestInstanceBasicMySQL(base_basic.TestInstanceBasicMySQLBase):
    datastore = 'mysql'


class TestInstanceBasicMariaDB(base_basic.TestInstanceBasicMySQLBase):
    datastore = 'mariadb'


class TestInstanceBasicPostgreSQL(base_basic.TestInstanceBasicBase):
    datastore = 'postgresql'
    create_user = False
    enable_root = True

    def get_config_value(self, ip, option):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/postgres')
        with utils.SQLClient(db_url) as db_client:
            cmd = f"SELECT setting FROM pg_settings WHERE name='{option}';"
            ret = db_client.pgsql_execute(cmd)
            rows = ret.fetchall()

        self.assertEqual(1, len(rows))
        return int(rows[0][0])

    @decorators.idempotent_id("b6c03cb6-f40f-11ea-a950-00224d6b7bc1")
    def test_configuration(self):
        # Default is 100
        create_values = {"max_connections": 101}
        update_values = {"max_connections": 102}
        self.configuration_test(create_values, update_values,
                                need_restart=True)



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import time

from oslo_log import log as logging
from tempest import config

from trove_tempest_plugin.tests import base as trove_base

CONF = config.CONF
LOG = logging.getLogger(__name__)


class TestReplicationBase(trove_base.BaseTroveTest):
    def insert_data_replication(self, *args, **kwargs):
        pass

    def verify_data_replication(self, *args, **kwargs):
        pass

    def insert_data_after_promote(self, *args, **kwargs):
        pass

    def verify_data_after_promote(self, *args, **kwargs):
        pass

    def create_database(self, name, **kwargs):
        pass

    def replication_test(self):
        # Insert data for primary
        LOG.info(f"Inserting data before creating replicas on "
                 f"{self.instance_ip}")
        self.insert_data_replication(self.instance_ip)

        # Create replica1
        LOG.info(f"Creating replica1 for instance {self.instance_id}")
        name = self.get_resource_name("replica-01")
        replica1 = self.create_instance(name, replica_of=self.instance_id,
                                        create_user=self.create_user)
        replica1_id = replica1['id']
        self.addCleanup(self.wait_for_instance_status, replica1_id,
                        need_delete=True, expected_status='DELETED')
        self.wait_for_instance_status(
            replica1_id,
            expected_op_status=["HEALTHY"],
            timeout=CONF.database.database_build_timeout * 2)
        replica1 = self.client.get_resource(
            "instances", replica1_id)['instance']
        replica1_ip = self.get_instance_ip(replica1)

        # Verify API response of primary
        ret = self.client.get_resource('instances', self.instance_id)
        self.assertIsNotNone(ret['instance'].get('replicas'))
        self.assertEqual(1, len(ret['instance']['replicas']))
        self.assertEqual(replica1_id,
                         ret['instance']['replicas'][0]['id'])

        # Verify API response of replica1
        ret = self.client.get_resource('instances', replica1_id)
        self.assertIsNotNone(ret['instance'].get('replica_of'))
        self.assertEqual(self.instance_id,
                         ret['instance']['replica_of']['id'])

        # Verify databases created in replica
        time.sleep(5)
        LOG.info(f"Getting databases on primary {self.instance_ip}"
                 f"({self.instance_id}) and replica {replica1_ip}"
                 f"({replica1_id})")
        primary_dbs = self.get_databases(self.instance_id, ip=self.instance_ip)
        replica_dbs = self.get_databases(replica1_id, ip=replica1_ip)
        self.assertEqual(len(primary_dbs), len(replica_dbs))

        # Create a new database in primary and verify in replica
        LOG.info(f"Creating database in instance {self.instance_id}")
        db_name = 'db_for_replication'
        self.create_database(db_name, ip=self.instance_ip)

        time.sleep(5)
        LOG.info(f"Getting databases on primary {self.instance_ip}"
                 f"({self.instance_id}) and replica {replica1_ip}"
                 f"({replica1_id})")
        new_primary_dbs = self.get_databases(self.instance_id,
                                             ip=self.instance_ip)
        new_replica1_dbs = self.get_databases(replica1_id, ip=replica1_ip)
        self.assertEqual(len(new_primary_dbs), len(new_replica1_dbs))
        self.assertGreater(len(new_replica1_dbs), len(replica_dbs))
        new_db_names = [db['name'] for db in new_replica1_dbs]
        self.assertIn(db_name, new_db_names)

        # In the zuul environment, the testing VM can not create the
        # replica2 vm due to the lack of memory.
        if not CONF.database.run_full_tests:
            return
        # Create replica2
        LOG.info(f"Creating replica2 for instance {self.instance_id}")
        name = self.get_resource_name("replica-02")
        replica2 = self.create_instance(name, replica_of=self.instance_id,
                                        create_user=self.create_user)
        replica2_id = replica2['id']
        self.addCleanup(self.wait_for_instance_status, replica2_id,
                        need_delete=True, expected_status='DELETED')
        self.wait_for_instance_status(
            replica2_id,
            expected_op_status=["HEALTHY"],
            timeout=CONF.database.database_build_timeout * 2)
        replica2 = self.client.get_resource(
            "instances", replica2_id)['instance']
        replica2_ip = self.get_instance_ip(replica2)

        # Verify API response of primary and replica2
        ret = self.client.get_resource('instances', self.instance_id)
        self.assertIsNotNone(ret['instance'].get('replicas'))
        self.assertEqual(2, len(ret['instance']['replicas']))
        replica_ids = [replica['id']
                       for replica in ret['instance']['replicas']]
        self.assertIn(replica1_id, replica_ids)
        self.assertIn(replica2_id, replica_ids)

        # Verify databases synced to replica2
        time.sleep(5)
        LOG.info(f"Getting databases on replica {replica2_ip}({replica2_id})")
        replica2_dbs = self.get_databases(replica2_id, ip=replica2_ip)
        replica2_db_names = [db['name'] for db in replica2_dbs]
        self.assertIn(db_name, replica2_db_names)

        # Verify data synchronization on replica1 and replica2
        LOG.info(f"Verifying data on replicas {replica1_ip} and {replica2_ip}")
        self.verify_data_replication(replica1_ip)
        self.verify_data_replication(replica2_ip)

        # Volume resize to primary
        LOG.info(f"Resizing volume for primary {self.instance_id} to 2G")
        req_body = {
            "resize": {
                "volume": {"size": 2}
            }
        }
        self.client.create_resource(f"instances/{self.instance_id}/action",
                                    req_body, expected_status_code=202,
                                    need_response=False)
        self.wait_for_instance_status(self.instance_id,
                                      expected_op_status=["HEALTHY"])
        self.wait_for_instance_status(replica1_id,
                                      expected_op_status=["HEALTHY"])
        self.wait_for_instance_status(replica2_id,
                                      expected_op_status=["HEALTHY"])

        # Verify the volumes of all the replicas are also resized to 2G
        replica1 = self.client.get_resource('instances', replica1_id)
        self.assertEqual(2, replica1['instance']['volume'].get('size', 0))
        replica2 = self.client.get_resource('instances', replica2_id)
        self.assertEqual(2, replica2['instance']['volume'].get('size', 0))

        # Promote replica1 to primary
        LOG.info(f"Promoting replica1 {replica1_id} to primary")
        promote_primary = {
            "promote_to_replica_source": {}
        }
        self.client.create_resource(f"instances/{replica1_id}/action",
                                    promote_primary, expected_status_code=202,
                                    need_response=False)
        self.wait_for_instance_status(replica1_id,
                                      expected_op_status=["HEALTHY"])

        # Make sure to delete replicas first for clean up, in case failure
        # happens when replica1 is still the primary.
        self.addCleanup(self.wait_for_instance_status, self.instance_id,
                        need_delete=True, expected_status='DELETED')
        self.addCleanup(self.wait_for_instance_status, replica2_id,
                        need_delete=True, expected_status='DELETED')

        # Verify API response of the new primary
        ret = self.client.get_resource('instances', replica1_id)
        self.assertIsNotNone(ret['instance'].get('replicas'))
        self.assertEqual(2, len(ret['instance']['replicas']))
        replica_ids = [replica['id']
                       for replica in ret['instance']['replicas']]
        self.assertCountEqual([self.instance_id, replica2_id], replica_ids)

        # Verify API response of replicas
        ret = self.client.get_resource('instances', replica2_id)
        self.assertIsNotNone(ret['instance'].get('replica_of'))
        self.assertEqual(replica1_id, ret['instance']['replica_of']['id'])

        # Insert data to new primary and verify in replicas
        LOG.info(f"Inserting data on new primary {replica1_ip}")
        self.insert_data_after_promote(replica1_ip)
        time.sleep(5)
        LOG.info(f"Verifying data on new replicas {self.instance_ip} and "
                 f"{replica2_ip}")
        self.verify_data_after_promote(self.instance_ip)
        self.verify_data_after_promote(replica2_ip)

        # Detach original primary from the replication cluster
        LOG.info(f"Detaching replica {self.instance_id} from the replication "
                 f"cluster")
        detach_replica = {
            "instance": {
                "replica_of": ""
            }
        }
        self.client.put_resource(f'/instances/{self.instance_id}',
                                 detach_replica)
        self.wait_for_instance_status(self.instance_id,
                                      expected_op_status=["HEALTHY"])

        # Verify original primary
        ret = self.client.get_resource('instances', self.instance_id)
        self.assertIsNone(ret['instance'].get('replicas'))
        self.assertIsNone(ret['instance'].get('replica_of'))

        # Rebuild test for replication cluster, now replica1 is the primary
        if CONF.database.rebuild_image_id:
            LOG.info(f"Rebuilding primary {replica1_id}")
            self.rebuild_instance(replica1_id, CONF.database.rebuild_image_id)
            LOG.info(f"Verifying data after rebuild {replica1_id}")
            self.verify_data_after_promote(replica1_ip)

            LOG.info(f"Rebuilding replica {replica2_id}")
            self.rebuild_instance(replica2_id, CONF.database.rebuild_image_id)
            LOG.info(f"Verifying data after rebuild {replica2_id}")
            self.verify_data_after_promote(replica2_ip)



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from oslo_log import log as logging
from tempest import config

from trove_tempest_plugin.tests import base as trove_base

LOG = logging.getLogger(__name__)
CONF = config.CONF


class TestBackupBase(trove_base.BaseTroveTest):
    @classmethod
    def insert_data(cls, *args, **kwargs):
        pass

    @classmethod
    def insert_data_inc(cls, *args, **kwargs):
        pass

    def verify_data(self, *args, **kwargs):
        pass

    def verify_data_inc(self, *args, **kwargs):
        pass

    @classmethod
    def resource_setup(cls):
        super(TestBackupBase, cls).resource_setup()

        # Trove will automatically create a swift container for backup. We need
        # to make sure there is no swift container left for test user after
        # testing.
        if CONF.database.remove_swift_account:
            cls.addClassResourceCleanup(cls.delete_swift_account)

        # Insert some data to the current db instance
        LOG.info(f"Inserting data on {cls.instance_ip} before creating full"
                 f"backup")
        cls.insert_data(cls.instance_ip)

        # Create a backup that is shared within this test class.
        LOG.info(f"Creating full backup for instance {cls.instance_id}")
        name = cls.get_resource_name("backup")
        backup = cls.create_backup(cls.instance_id, name)
        cls.wait_for_backup_status(backup['id'])
        cls.backup = cls.client.get_resource("backups", backup['id'])['backup']

    def backup_full_test(self):
        # Restore from backup
        LOG.info(f'Creating a new instance using the backup '
                 f'{self.backup["id"]}')
        name = self.get_resource_name("restore")
        restore_instance = self.create_instance(
            name,
            datastore_version=self.backup['datastore']['version'],
            backup_id=self.backup['id'],
            create_user=self.create_user
        )
        self.wait_for_instance_status(
            restore_instance['id'],
            expected_op_status=["HEALTHY"],
            timeout=CONF.database.database_restore_timeout)

        if self.enable_root:
            self.root_password = self.get_root_pass(restore_instance['id'])

        restore_instance = self.client.get_resource(
            "instances", restore_instance['id'])['instance']
        restore_instance_ip = self.get_instance_ip(restore_instance)

        LOG.info(f"Verifying data on restored instance {restore_instance_ip}")
        self.verify_data(restore_instance_ip)

        # Delete the new instance explicitly to avoid too many instances
        # during the test.
        self.wait_for_instance_status(restore_instance['id'],
                                      expected_status="DELETED",
                                      need_delete=True)

    def backup_incremental_test(self):
        # Insert some data
        LOG.info(f"Inserting data on {self.instance_ip} before creating "
                 f"incremental backup")
        self.insert_data_inc(self.instance_ip)

        # Create a second backup
        LOG.info(f"Creating an incremental backup based on "
                 f"{self.backup['id']}")
        name = self.get_resource_name("backup-inc")
        backup_inc = self.create_backup(
            self.instance_id, name, incremental=True,
            parent_id=self.backup['id']
        )
        self.wait_for_backup_status(backup_inc['id'])

        # Restore from backup
        LOG.info(f"Creating a new instance using the backup "
                 f"{backup_inc['id']}")
        name = self.get_resource_name("restore-inc")
        restore_instance = self.create_instance(
            name,
            datastore_version=backup_inc['datastore']['version'],
            backup_id=backup_inc['id'],
            create_user=self.create_user
        )
        self.wait_for_instance_status(
            restore_instance['id'],
            expected_op_status=["HEALTHY"],
            timeout=CONF.database.database_restore_timeout)

        if self.enable_root:
            self.root_password = self.get_root_pass(restore_instance['id'])

        restore_instance = self.client.get_resource(
            "instances", restore_instance['id'])['instance']
        restore_instance_ip = self.get_instance_ip(restore_instance)

        LOG.info(f"Verifying data on {restore_instance_ip}"
                 f"({restore_instance['id']}) after restoring incremental "
                 f"backup")
        self.verify_data_inc(restore_instance_ip)

        # Delete the new instance explicitly to avoid too many instances
        # during the test.
        self.wait_for_instance_status(restore_instance['id'],
                                      expected_status="DELETED",
                                      need_delete=True)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import time

from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import netutils
from oslo_utils import uuidutils
from swiftclient import client as swift_client
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions
from tempest import test
import tenacity

from trove_tempest_plugin.tests import constants
from trove_tempest_plugin.tests import exceptions as trove_exc
from trove_tempest_plugin.tests import utils

CONF = config.CONF
LOG = logging.getLogger(__name__)


class BaseTroveTest(test.BaseTestCase):
    credentials = ('admin', 'primary')
    datastore = None
    instance = None
    instance_id = None
    instance_ip = None
    password = ""
    create_user = True
    enable_root = False

    @classmethod
    def get_resource_name(cls, resource_type):
        prefix = "trove-tempest-%s" % cls.__name__
        return data_utils.rand_name(resource_type, prefix=prefix)

    @classmethod
    def skip_checks(cls):
        super(BaseTroveTest, cls).skip_checks()

        if not CONF.service_available.trove:
            raise cls.skipException("Database service is not available.")

        if cls.datastore not in CONF.database.enabled_datastores:
            raise cls.skipException(
                "Datastore %s is not enabled." % cls.datastore
            )

    @classmethod
    def get_swift_client(cls):
        auth_version = "3.0"
        auth_url = CONF.identity.uri_v3
        user = cls.os_primary.credentials.username
        key = cls.os_primary.credentials.password
        tenant_name = cls.os_primary.credentials.project_name
        region_name = cls.os_primary.region
        os_options = {'tenant_name': tenant_name, 'region_name': region_name}

        return swift_client.Connection(
            auth_url, user, key, auth_version=auth_version,
            os_options=os_options)

    @classmethod
    def get_swift_admin_client(cls):
        auth_version = "3.0"
        auth_url = CONF.identity.uri_v3
        user = cls.os_admin.credentials.username
        key = cls.os_admin.credentials.password
        tenant_name = cls.os_admin.credentials.project_name
        region_name = cls.os_admin.region
        os_options = {'tenant_name': tenant_name, 'region_name': region_name}

        return swift_client.Connection(
            auth_url, user, key, auth_version=auth_version,
            os_options=os_options)

    @classmethod
    def setup_clients(cls):
        super(BaseTroveTest, cls).setup_clients()

        cls.client = cls.os_primary.database.TroveClient()
        cls.admin_client = cls.os_admin.database.TroveClient()
        cls.admin_server_client = cls.os_admin.servers_client
        cls.account_client = cls.os_primary.account_client
        cls.container_client = cls.os_primary.container_client
        cls.object_client = cls.os_primary.object_client
        cls.admin_container_client = cls.os_admin.container_client
        cls.admin_object_client = cls.os_admin.object_client
        # Swift client is special, we want to re-use the log_generator func
        # in python-troveclient.
        cls.swift = cls.get_swift_client()
        cls.swift_admin = cls.get_swift_admin_client()

    @classmethod
    def setup_credentials(cls):
        # Do not create network resources automatically.
        cls.set_network_resources()
        super(BaseTroveTest, cls).setup_credentials()

    @classmethod
    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(exceptions.Conflict),
        wait=tenacity.wait_incrementing(1, 1, 5),
        stop=tenacity.stop_after_attempt(15)
    )
    def _delete_network(cls, net_id):
        """Make sure the network is deleted.

        Neutron can be slow to clean up ports from the subnets/networks.
        Retry this delete a few times if we get a "Conflict" error to give
        neutron time to fully cleanup the ports.
        """
        networks_client = cls.os_primary.networks_client
        try:
            networks_client.delete_network(net_id)
        except Exception:
            LOG.error('Unable to delete network %s', net_id)
            raise

    @classmethod
    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(exceptions.Conflict),
        wait=tenacity.wait_incrementing(1, 1, 5),
        stop=tenacity.stop_after_attempt(15)
    )
    def _delete_subnet(cls, subnet_id):
        """Make sure the subnet is deleted.

        Neutron can be slow to clean up ports from the subnets/networks.
        Retry this delete a few times if we get a "Conflict" error to give
        neutron time to fully cleanup the ports.
        """
        subnets_client = cls.os_primary.subnets_client
        try:
            subnets_client.delete_subnet(subnet_id)
        except Exception:
            LOG.error('Unable to delete subnet %s', subnet_id)
            raise

    @classmethod
    def _create_network(cls):
        """Create database instance network."""
        networks_client = cls.os_primary.networks_client
        subnets_client = cls.os_primary.subnets_client
        routers_client = cls.os_primary.routers_client

        if CONF.database.shared_network:
            private_network = CONF.database.shared_network
            if not uuidutils.is_uuid_like(private_network):
                networks = networks_client.list_networks()['networks']
                for net in networks:
                    if net['name'] == private_network:
                        private_network = net['id']
                        break
                else:
                    raise exceptions.NotFound(
                        'Shared network %s not found' % private_network
                    )

            cls.private_network = private_network
            return

        network_kwargs = {"name": cls.get_resource_name("network")}
        result = networks_client.create_network(**network_kwargs)
        LOG.info('Private network created: %s', result['network'])
        cls.private_network = result['network']["id"]
        cls.addClassResourceCleanup(
            utils.wait_for_removal,
            cls._delete_network,
            networks_client.show_network,
            cls.private_network
        )

        subnet_kwargs = {
            'name': cls.get_resource_name("subnet"),
            'network_id': cls.private_network,
            'cidr': CONF.database.subnet_cidr,
            'ip_version': 4
        }
        result = subnets_client.create_subnet(**subnet_kwargs)
        subnet_id = result['subnet']['id']
        LOG.info('Private subnet created: %s', result['subnet'])
        cls.addClassResourceCleanup(
            utils.wait_for_removal,
            cls._delete_subnet,
            subnets_client.show_subnet,
            subnet_id
        )

        # In dev node, Trove instance needs to connect with control host
        router_params = {
            'name': cls.get_resource_name("router"),
            'external_gateway_info': {
                "network_id": CONF.network.public_network_id
            }
        }
        result = routers_client.create_router(**router_params)
        router_id = result['router']['id']
        LOG.info('Private router created: %s', result['router'])
        cls.addClassResourceCleanup(
            utils.wait_for_removal,
            routers_client.delete_router,
            routers_client.show_router,
            router_id
        )

        routers_client.add_router_interface(router_id, subnet_id=subnet_id)
        LOG.info('Subnet %s added to the router %s', subnet_id, router_id)
        cls.addClassResourceCleanup(
            routers_client.remove_router_interface,
            router_id,
            subnet_id=subnet_id
        )

    @classmethod
    def resource_setup(cls):
        super(BaseTroveTest, cls).resource_setup()

        # Create network for database instance, use cls.private_network as the
        # network ID.
        cls._create_network()

        instance = cls.create_instance(create_user=cls.create_user)
        cls.instance_id = instance['id']
        cls.wait_for_instance_status(cls.instance_id,
                                     expected_op_status=["HEALTHY"])
        cls.instance = cls.client.get_resource(
            "instances", cls.instance_id)['instance']
        cls.instance_ip = cls.get_instance_ip(cls.instance)

        if cls.enable_root:
            cls.password = cls.get_root_pass(cls.instance_id)

    def assert_single_item(self, items, **props):
        return self.assert_multiple_items(items, 1, **props)[0]

    def assert_multiple_items(self, items, count, **props):
        """Check if a object is in a list of objects.

        e.g. props is a sub-dict, items is a list of dicts.
        """

        def _matches(item, **props):
            for prop_name, prop_val in props.items():
                v = item[prop_name] if isinstance(
                    item, dict) else getattr(item, prop_name)

                if v != prop_val:
                    return False

            return True

        filtered_items = list(
            [item for item in items if _matches(item, **props)]
        )

        found = len(filtered_items)

        if found != count:
            LOG.info("[FAIL] items=%s, expected_props=%s", str(items), props)
            self.fail("Wrong number of items found [props=%s, "
                      "expected=%s, found=%s]" % (props, count, found))

        return filtered_items

    @classmethod
    def delete_swift_account(cls):
        LOG.info(f"Cleaning up Swift account")
        try:
            cls.account_client.delete('')
        except exceptions.NotFound:
            pass

    @classmethod
    def delete_swift_containers(cls, container_client, object_client,
                                containers):
        """Remove containers and all objects in them.

        The containers should be visible from the container_client given.
        Will not throw any error if the containers don't exist.
        Will not check that object and container deletions succeed.
        After delete all the objects from a container, it will wait 3
        seconds before delete the container itself, in order for deployments
        using HA proxy sync the deletion properly, otherwise, the container
        might fail to be deleted because it's not empty.
        """
        if isinstance(containers, str):
            containers = [containers]

        for cont in containers:
            try:
                params = {'limit': 9999, 'format': 'json'}
                _, objlist = container_client.list_container_objects(
                    cont,
                    params)
                # delete every object in the container
                for obj in objlist:
                    test_utils.call_and_ignore_notfound_exc(
                        object_client.delete_object, cont, obj['name'])

                # sleep 3 seconds to sync the deletion of the objects
                # in HA deployment
                time.sleep(3)

                container_client.delete_container(cont)
            except exceptions.NotFound:
                pass

    @classmethod
    def create_instance(cls, name=None, datastore_version=None,
                        database=constants.DB_NAME, username=constants.DB_USER,
                        password=constants.DB_PASS, backup_id=None,
                        replica_of=None, create_user=True):
        """Create database instance.

        Creating database instance is time-consuming, so we define this method
        as a class method, which means the instance is shared in a single
        TestCase. According to
        https://docs.openstack.org/tempest/latest/write_tests.html#adding-a-new-testcase,
        all test methods within a TestCase are assumed to be executed serially.
        """
        name = name or cls.get_resource_name("instance")

        # Flavor, volume, datastore are not needed for creating replica.
        if replica_of:
            body = {
                "instance": {
                    "name": name,
                    "nics": [{"net-id": cls.private_network}],
                    "access": {"is_public": True},
                    "replica_of": replica_of,
                }
            }

        # Get datastore version. Get from API if the default ds version is not
        # configured.
        elif not datastore_version:
            default_versions = CONF.database.default_datastore_versions
            datastore_version = default_versions.get(cls.datastore)
            if not datastore_version:
                res = cls.client.list_resources("datastores")
                for d in res['datastores']:
                    if d['name'] == cls.datastore:
                        if d.get('default_version'):
                            datastore_version = d['default_version']
                        else:
                            datastore_version = d['versions'][0]['name']
                        break
            if not datastore_version:
                message = ('Failed to get available datastore version.')
                raise exceptions.TempestException(message)

        if not replica_of:
            body = {
                "instance": {
                    "name": name,
                    "datastore": {
                        "type": cls.datastore,
                        "version": datastore_version
                    },
                    "flavorRef": CONF.database.flavor_id,
                    "volume": {
                        "size": 1,
                        "type": CONF.database.volume_type
                    },
                    "nics": [{"net-id": cls.private_network}],
                    "access": {"is_public": True}
                }
            }
            if backup_id:
                body['instance'].update(
                    {'restorePoint': {'backupRef': backup_id}})
            if create_user:
                body['instance'].update({
                    'databases': [{"name": database}],
                    "users": [
                        {
                            "name": username,
                            "password": password,
                            "databases": [{"name": database}]
                        }
                    ]
                })

        res = cls.client.create_resource("instances", body)
        cls.addClassResourceCleanup(cls.wait_for_instance_status,
                                    res["instance"]["id"],
                                    need_delete=True,
                                    expected_status="DELETED")

        return res["instance"]

    @classmethod
    def restart_instance(cls, instance_id):
        """Restart database service and wait until it's healthy."""
        cls.client.create_resource(
            f"instances/{instance_id}/action",
            {"restart": {}},
            expected_status_code=202,
            need_response=False)
        cls.wait_for_instance_status(instance_id,
                                     expected_op_status=["HEALTHY"])

    @classmethod
    def wait_for_instance_status(cls, id,
                                 expected_status=["ACTIVE"],
                                 expected_op_status=[],
                                 need_delete=False,
                                 timeout=CONF.database.database_build_timeout):
        def _wait():
            try:
                res = cls.client.get_resource("instances", id)
                cur_status = res["instance"]["status"]
            except exceptions.NotFound:
                if need_delete or "DELETED" in expected_status:
                    LOG.info('Instance %s is deleted', id)
                    raise loopingcall.LoopingCallDone()
                return

            if cur_status in expected_status:
                LOG.info('Instance %s becomes %s', id, cur_status)
                if expected_op_status:
                    op_status = res["instance"]["operating_status"]
                    if op_status in expected_op_status:
                        raise loopingcall.LoopingCallDone()
                else:
                    raise loopingcall.LoopingCallDone()

            elif "ERROR" not in expected_status and cur_status == "ERROR":
                # If instance status goes to ERROR but is not expected, stop
                # waiting

                res = cls.admin_client.get_resource("instances", id)
                LOG.info(f'Instance fault msg: {res["instance"].get("fault")}')

                # Show trove-guestagent log for debug purpose.
                # Only admin user is able to publish and show the trove guest
                # agent log. Make sure the container is deleted after fetching
                # the log.
                try:
                    LOG.info(f"Publishing guest log for instance {id}")
                    cls.publish_log(id, 'guest')
                    LOG.info(f"Getting guest log content for instance {id}")
                    log_gen = cls.log_generator(id, 'guest', lines=0)
                    log_content = "".join([chunk for chunk in log_gen()])
                    LOG.info(
                        f"\n=============================================\n"
                        f"Trove guest agent log for instance {id}\n"
                        f"=============================================")
                    LOG.info(log_content)
                except Exception as err:
                    LOG.warning(f"Failed to get guest log for instance {id}, "
                                f"error: {str(err)}")
                finally:
                    # Remove the swift container of database logs.
                    LOG.info(f"Deleting swift container "
                             f"{CONF.database.database_log_container}")
                    cls.delete_swift_containers(
                        cls.admin_container_client, cls.admin_object_client,
                        CONF.database.database_log_container)

                message = "Instance status is ERROR."
                caller = test_utils.find_test_caller()
                if caller:
                    message = '({caller}) {message}'.format(caller=caller,
                                                            message=message)
                raise exceptions.UnexpectedResponseCode(message)

        if type(expected_status) != list:
            expected_status = [expected_status]

        if type(expected_op_status) != list:
            expected_op_status = [expected_op_status]

        if need_delete:
            # If resource already removed, return
            try:
                cls.client.get_resource("instances", id)
            except exceptions.NotFound:
                LOG.info('Instance %s not found', id)
                return

            LOG.info(f"Deleting instance {id}")
            cls.admin_client.force_delete_instance(id)

        timer = loopingcall.FixedIntervalWithTimeoutLoopingCall(_wait)
        try:
            timer.start(interval=10, timeout=timeout, initial_delay=5).wait()
        except loopingcall.LoopingCallTimeOut:
            message = ("Instance %s is not in the expected status: %s" %
                       (id, expected_status))
            caller = test_utils.find_test_caller()
            if caller:
                message = '({caller}) {message}'.format(caller=caller,
                                                        message=message)
            raise exceptions.TimeoutException(message)

    @classmethod
    def get_instance_ip(cls, instance=None):
        if not instance:
            instance = cls.client.get_resource(
                "instances", cls.instance_id)['instance']

        # TODO(lxkong): IPv6 needs to be tested.
        v4_ip = None

        if 'addresses' in instance:
            for addr_info in instance['addresses']:
                if addr_info['type'] == 'private':
                    v4_ip = addr_info['address']
                if addr_info['type'] == 'public':
                    v4_ip = addr_info['address']
                    break
        else:
            ips = instance.get('ip', [])
            for ip in ips:
                if netutils.is_valid_ipv4(ip):
                    v4_ip = ip

        if not v4_ip:
            message = ('Failed to get instance IP address.')
            raise exceptions.TempestException(message)

        return v4_ip

    def get_databases(self, instance_id, **kwargs):
        url = f'instances/{instance_id}/databases'
        ret = self.client.list_resources(url)
        return ret['databases']

    def get_users(self, instance_id):
        url = f'instances/{instance_id}/users'
        ret = self.client.list_resources(url)
        return ret['users']

    @classmethod
    def create_backup(cls, instance_id, backup_name, incremental=False,
                      parent_id=None, description=None):
        body = {
            "backup": {
                "name": backup_name,
                "instance": instance_id,
                "incremental": 1 if incremental else 0,
            }
        }
        if description:
            body['backup']['description'] = description
        if parent_id:
            body['backup']['parent_id'] = parent_id

        res = cls.client.create_resource("backups", body,
                                         expected_status_code=202)
        cls.addClassResourceCleanup(cls.wait_for_backup_status,
                                    res["backup"]['id'],
                                    expected_status='',
                                    need_delete=True)
        return res["backup"]

    @classmethod
    def delete_backup(cls, backup_id, ignore_notfound=False):
        cls.client.delete_resource('backups', backup_id,
                                   ignore_notfound=ignore_notfound)

    @classmethod
    def wait_for_backup_status(cls, id, expected_status=["COMPLETED"],
                               need_delete=False):
        def _wait():
            try:
                res = cls.client.get_resource("backups", id)
                cur_status = res["backup"]["status"]
            except exceptions.NotFound:
                if need_delete or "DELETED" in expected_status:
                    LOG.info('Backup %s is deleted', id)
                    raise loopingcall.LoopingCallDone()
                return

            if cur_status in expected_status:
                LOG.info('Backup %s becomes %s', id, cur_status)
                raise loopingcall.LoopingCallDone()
            elif "FAILED" not in expected_status and cur_status == "FAILED":
                # If backup status goes to FAILED but is not expected, stop
                # waiting
                message = "Backup status is FAILED."
                caller = test_utils.find_test_caller()
                if caller:
                    message = '({caller}) {message}'.format(caller=caller,
                                                            message=message)
                raise exceptions.UnexpectedResponseCode(message)

        if type(expected_status) != list:
            expected_status = [expected_status]

        if need_delete:
            # If resource already removed, return
            try:
                cls.client.get_resource("backups", id)
            except exceptions.NotFound:
                LOG.info('Backup %s not found', id)
                return

            LOG.info(f"Deleting backup {id}")
            cls.delete_backup(id, ignore_notfound=True)

        timer = loopingcall.FixedIntervalWithTimeoutLoopingCall(_wait)
        try:
            timer.start(interval=10,
                        timeout=CONF.database.backup_wait_timeout).wait()
        except loopingcall.LoopingCallTimeOut:
            message = ("Backup %s is not in the expected status: %s" %
                       (id, expected_status))
            caller = test_utils.find_test_caller()
            if caller:
                message = '({caller}) {message}'.format(caller=caller,
                                                        message=message)
            raise exceptions.TimeoutException(message)

    @classmethod
    def get_root_pass(cls, instance_id):
        resp = cls.client.create_resource(f"instances/{instance_id}/root", {})
        return resp['user']['password']

    @classmethod
    def rebuild_instance(cls, instance_id, image_id):
        rebuild_req = {
            "rebuild": {
                "image_id": image_id
            }
        }
        cls.admin_client.create_resource(
            f"mgmt/instances/{instance_id}/action",
            rebuild_req, expected_status_code=202,
            need_response=False)
        cls.wait_for_instance_status(instance_id,
                                     expected_op_status=["HEALTHY"])

    @classmethod
    def create_config(cls, name, values, datastore, datastore_version):
        create_config = {
            "configuration": {
                "datastore": {
                    "type": datastore,
                    "version": datastore_version
                },
                "values": values,
                "name": name
            }
        }
        config = cls.client.create_resource('configurations', create_config)
        return config

    @classmethod
    def attach_config(cls, instance_id, config_id):
        attach_config = {
            "instance": {
                "configuration": config_id
            }
        }
        cls.client.put_resource(f'instances/{instance_id}', attach_config)

    @classmethod
    def detach_config(cls, instance_id):
        detach_config = {
            "instance": {
                "configuration": None
            }
        }
        cls.client.put_resource(f'instances/{instance_id}', detach_config)

    @classmethod
    def publish_log(cls, instance_id, name='guest'):
        client = cls.admin_client if name == 'guest' else cls.client
        req_body = {
            'name': name,
            'publish': 1
        }
        client.create_resource(f"instances/{instance_id}/log",
                               req_body)

    @classmethod
    def get_log_info(cls, instance_id, name='guest'):
        req_body = {
            'name': name,
        }
        return cls.admin_client.create_resource(
            f"instances/{instance_id}/log",
            req_body)

    @classmethod
    def _get_container_info(cls, instance_id, log_name):
        try:
            log_info = cls.get_log_info(instance_id, log_name)['log']
            container = log_info['container']
            prefix = log_info['prefix']
            metadata_file = log_info['metafile']
            return container, prefix, metadata_file
        except swift_client.ClientException as ex:
            if ex.http_status == 404:
                raise trove_exc.GuestLogNotFound()
            raise trove_exc.TroveTempestException()

    @classmethod
    def log_generator(cls, instance_id, log_name, lines=50):
        """Return generator to yield the last <lines> lines of guest log.

        This method is copied from python-troveclient.
        """
        swift_cli = cls.swift_admin if log_name == 'guest' else cls.swift

        def _log_generator(instance_id, log_name, lines):
            try:
                container, prefix, metadata_file = cls._get_container_info(
                    instance_id, log_name)

                head, body = swift_cli.get_container(container, prefix=prefix)
                log_obj_to_display = []

                if lines:
                    total_lines = lines
                    partial_results = False
                    parts = sorted(body, key=lambda obj: obj['last_modified'],
                                   reverse=True)

                    for part in parts:
                        obj_hdrs = swift_cli.head_object(
                            container,
                            part['name'])
                        obj_lines = int(obj_hdrs['x-object-meta-lines'])
                        log_obj_to_display.insert(0, part)
                        if obj_lines >= lines:
                            partial_results = True
                            break
                        lines -= obj_lines
                    if not partial_results:
                        lines = total_lines

                    part = log_obj_to_display.pop(0)
                    hdrs, log_obj = swift_cli.get_object(
                        container,
                        part['name'])
                    log_by_lines = log_obj.decode().splitlines()
                    yield "\n".join(log_by_lines[-1 * lines:]) + "\n"
                else:
                    # Show all the logs
                    log_obj_to_display = sorted(
                        body, key=lambda obj: obj['last_modified'])

                for log_part in log_obj_to_display:
                    headers, log_obj = swift_cli.get_object(
                        container,
                        log_part['name'])
                    yield log_obj.decode()
            except swift_client.ClientException as ex:
                if ex.http_status == 404:
                    raise trove_exc.GuestLogNotFound()
                raise trove_exc.TroveTempestException()

        return lambda: _log_generator(instance_id, log_name, lines)



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from tempest.lib import decorators

from trove_tempest_plugin.tests import constants
from trove_tempest_plugin.tests.scenario import base_replication
from trove_tempest_plugin.tests import utils


class TestReplicationMySQL(base_replication.TestReplicationBase):
    datastore = 'mysql'

    def insert_data_replication(self, ip,
                                username=constants.DB_USER,
                                password=constants.DB_PASS,
                                database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "CREATE TABLE Persons (ID int, String varchar(255));",
                "insert into Persons VALUES (1, 'replication');"
            ]
            db_client.mysql_execute(cmds)

    def verify_data_replication(self, ip,
                                username=constants.DB_USER,
                                password=constants.DB_PASS,
                                database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Persons;"
            ret = db_client.mysql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertEqual(1, len(rows))

        result = []
        for index in range(len(rows)):
            result.append(dict(zip(keys, rows[index])))
        expected = {'ID': 1, 'String': 'replication'}
        self.assert_single_item(result, **expected)

    def insert_data_after_promote(self, ip,
                                  username=constants.DB_USER,
                                  password=constants.DB_PASS,
                                  database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "insert into Persons VALUES (2, 'promote');"
            ]
            db_client.mysql_execute(cmds)

    def verify_data_after_promote(self, ip,
                                  username=constants.DB_USER,
                                  password=constants.DB_PASS,
                                  database=constants.DB_NAME):
        db_url = f'mysql+pymysql://{username}:{password}@{ip}:3306/{database}'
        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Persons;"
            ret = db_client.mysql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertGreater(len(rows), 1)

        result = []
        for index in range(len(rows)):
            result.append(dict(zip(keys, rows[index])))
        expected = {'ID': 2, 'String': 'promote'}
        self.assert_single_item(result, **expected)

    def create_database(self, name, **kwargs):
        create_db = {"databases": [{"name": name}]}
        self.client.create_resource(f"instances/{self.instance_id}/databases",
                                    create_db, expected_status_code=202,
                                    need_response=False)

    @decorators.idempotent_id("280d09c6-b027-11ea-b87c-00224d6b7bc1")
    def test_replication(self):
        self.replication_test()


class TestReplicationPostgreSQL(base_replication.TestReplicationBase):
    datastore = 'postgresql'
    create_user = False
    enable_root = True

    def insert_data_replication(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/postgres')

        with utils.SQLClient(db_url) as db_client:
            cmd = "CREATE DATABASE testdb;"
            db_client.pgsql_execute(cmd)

        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "CREATE TABLE Persons (ID int, String varchar(255));",
                "insert into Persons VALUES (1, 'replication');"
            ]
            db_client.pgsql_execute(cmds)

    def verify_data_replication(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/testdb')

        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Persons;"
            ret = db_client.pgsql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertEqual(1, len(rows))

        result = []
        for index in range(len(rows)):
            result.append(dict(zip(keys, rows[index])))
        expected = {'id': 1, 'string': 'replication'}
        self.assert_single_item(result, **expected)

    def insert_data_after_promote(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/testdb')
        with utils.SQLClient(db_url) as db_client:
            cmds = [
                "insert into Persons VALUES (2, 'promote');"
            ]
            db_client.pgsql_execute(cmds)

    def verify_data_after_promote(self, ip):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/testdb')

        with utils.SQLClient(db_url) as db_client:
            cmd = "select * from Persons;"
            ret = db_client.pgsql_execute(cmd)
            keys = ret.keys()
            rows = ret.fetchall()
        self.assertGreater(len(rows), 1)

        result = []
        for index in range(len(rows)):
            result.append(dict(zip(keys, rows[index])))
        expected = {'id': 2, 'string': 'promote'}
        self.assert_single_item(result, **expected)

    def get_databases(self, instance_id, ip="", **kwargs):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/postgres')

        with utils.SQLClient(db_url) as db_client:
            cmd = "SELECT datname FROM pg_catalog.pg_database WHERE " \
                  "(datistemplate ISNULL OR datistemplate = false);"
            ret = db_client.pgsql_execute(cmd)
            rows = ret.fetchall()

        dbs = []
        for row in rows:
            dbs.append({'name': row[0]})

        return dbs

    def create_database(self, name, ip=""):
        db_url = (f'postgresql+psycopg2://root:{self.password}@'
                  f'{ip}:5432/postgres')

        with utils.SQLClient(db_url) as db_client:
            cmd = f"CREATE DATABASE {name};"
            db_client.pgsql_execute(cmd)

    @decorators.idempotent_id("2f37f064-f418-11ea-a950-00224d6b7bc1")
    def test_replication(self):
        self.replication_test()


class TestReplicationMariaDB(TestReplicationMySQL):
    datastore = 'mariadb'



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from tempest.lib import decorators

from trove_tempest_plugin.tests import base as trove_base

LOG = logging.getLogger(__name__)


class TestInstanceUpdateAccessBase(trove_base.BaseTroveTest):
    """Update instance access base class.

    Updating instance access needs to change the public IP address of the
    instance, so we need a separate test class for this.
    """
    def update_access_test(self):
        """Test update instance accessbility"""
        if 'access' not in self.instance:
            raise self.skipException("Access not supported in API.")

        # Change instance to be private
        LOG.info(f"Changing instance {self.instance_id} to be private")
        body = {
            "instance": {
                "access": {
                    "is_public": False,
                }
            }
        }
        self.client.put_resource(f'instances/{self.instance_id}', body)
        self.wait_for_instance_status(self.instance_id,
                                      expected_op_status=["HEALTHY"],
                                      timeout=30)

        instance = self.client.get_resource(
            "instances", self.instance_id)['instance']
        self.assertFalse(instance['access']['is_public'])
        types = [addr['type'] for addr in instance['addresses']]
        self.assertNotIn('public', types)

        # Change back to public
        LOG.info(f"Changing instance {self.instance_id} to be public")
        body = {
            "instance": {
                "access": {
                    "is_public": True,
                }
            }
        }
        self.client.put_resource(f'instances/{self.instance_id}', body)
        self.wait_for_instance_status(self.instance_id,
                                      expected_op_status=["HEALTHY"],
                                      timeout=30)

    @decorators.idempotent_id("c907cc80-36b4-11eb-b177-00224d6b7bc1")
    def test_instance_update_access(self):
        self.update_access_test()
